//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                          //
//                                           Marblevator, Golf                                              //
//                                                                                                          //
//                                          Copyright © 2022 by                                             //
//                                        Zumwalt Properties, LLC.                                          //
//                                          All Rights Reserved.                                            //
//                                                                                                          //
//          This software is for NON COMMERCIAL USE ONLY in accordance with the licensing agreement         //
//                                     as published on this website.                                        //
//                                                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Includes.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <Servo.h>

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Constants.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // General.

  #define                           SERVO_HOME                0
  
  // Senior.
  
    // Head servo.
    
    #define                         HEAD_ANGLE_INIT           90
    #define                         HEAD_ANGLE_MAX            60
    #define                         HEAD_SERVO_PIN            6
    
    // Hip servo.
    
    #define                         HIP_ANGLE_INIT            0
    #define                         HIP_ANGLE_MAX             180
    #define                         HIP_SERVO_PIN             7
    
    // Lift servo.
    
    #define                         LIFT_ANGLE_INIT           0
    #define                         LIFT_ANGLE_MAX            180
    #define                         LIFT_SERVO_PIN            8
    
    // Shoulder servo.
    
    #define                         SHOULDER_ANGLE_INIT       90
    #define                         SHOULDER_ANGLE_MAX        80
    #define                         SHOULDER_SERVO_PIN        9
    
  // Junior.
  
    // Arm servo.
    
    #define                         JUNIOR_ARM_ANGLE_INIT     145
    #define                         JUNIOR_ARM_ANGLE_MAX      JUNIOR_ARM_ANGLE_INIT + 30
    #define                         JUNIOR_ARM_SERVO_PIN      10
    
    // Head servo.
    
    #define                         JUNIOR_HEAD_ANGLE_END     150
    #define                         JUNIOR_HEAD_ANGLE_INIT    40
    #define                         JUNIOR_HEAD_ANGLE_MAX     100
    #define                         JUNIOR_HEAD_SERVO_PIN     11

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Global variables.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // Senior.
  
    // Head servo.
    
    int                             nHead = HEAD_ANGLE_INIT;
    Servo                           servoHead;
    int                             nHeadDirection = 1;
  
    // Hip servo.
    
    int                             nHip = HIP_ANGLE_INIT;
    Servo                           servoHip;
    int                             nHipDirection = 1;
  
    // Lift servo.
    
    int                             nLift = LIFT_ANGLE_INIT;
    Servo                           servoLift;
    int                             nLiftDirection = 1;
  
    // Shoulder servo.
    
    int                             nShoulder = SHOULDER_ANGLE_INIT;
    Servo                           servoShoulder;
    int                             nShoulderDirection = 1;

  // Junior.
  
    // Arm servo.
    
    int                             nJuniorArm = JUNIOR_ARM_ANGLE_INIT;
    Servo                           servoJuniorArm;
    int                             nJuniorArmDirection = 1;
    
    // Head servo.
    
    int                             nJuniorHead = JUNIOR_HEAD_ANGLE_INIT;
    Servo                           servoJuniorHead;
    int                             nJuniorHeadDirection = 1;
    
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
//  setup.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void setup() 
{
  // Serial.
  
  Serial.begin(115200);

  // Attach servos.

  servoLift.attach(LIFT_SERVO_PIN);
  servoHead.attach(HEAD_SERVO_PIN);
  servoHip.attach(HIP_SERVO_PIN);
  servoShoulder.attach(SHOULDER_SERVO_PIN);
  servoJuniorArm.attach(JUNIOR_ARM_SERVO_PIN);
  servoJuniorHead.attach(JUNIOR_HEAD_SERVO_PIN);

  // Initialize servo positions.

  servoLift.write(nLift);
  servoHead.write(nHead);
  servoShoulder.write(nShoulder);
  servoHip.write(nHip);
  servoJuniorArm.write(nJuniorArm);
  servoJuniorHead.write(nJuniorHead);

  // If servo home is true, then loop here forever.
  
  while(SERVO_HOME){};
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// loop.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void loop() 
{
  // Loop forever.
  
  while(1)
  {
    // Prepare to putt.
  
    PrepareToPutt();
    
    // Putt.

    Serial.print("Putt...\n  Put #1.\n");
    
    Putt();
  
    // Perform "nPutts" subsequent putts.
  
      // Pick a random number from 0 through 9.
      
      int nPutts = random(0, 10);
  
      // Putt nPutts times.
      
      for(int nTemp = 0; nTemp < nPutts; nTemp ++)
      {
        // Raise the ball.
      
        Serial.print("  Raise the ball.\n");
        while(nLift <= LIFT_ANGLE_INIT + LIFT_ANGLE_MAX - 50)
        {
          nLift ++;
          servoLift.write(nLift);
          delayMicroseconds(1500);
        }

        // Display the put number.
        
        char  chBuffer[81];
        sprintf(chBuffer, "  Put #%d.\n", nTemp + 2);
        Serial.print(chBuffer);
        
        // Forward swing.
  
        ForwardSwing();
        
        // Rapidly lower the lift for the ball return.
        
        servoLift.write(nLift = LIFT_ANGLE_INIT);
        
        // Move the heads to follow the putt.
  
        HeadsFollowBall();
        
        // Prepare for the  next swing.
    
        if(nTemp < nPutts - 1)
        {
          Serial.print("  Back swing.\n");
          
          while((nShoulder > SHOULDER_ANGLE_INIT - SHOULDER_ANGLE_MAX) || 
                (nHead < HEAD_ANGLE_INIT) ||
                (nJuniorHead > 90))
          {
            if(nShoulder > SHOULDER_ANGLE_INIT - SHOULDER_ANGLE_MAX)
            {
              nShoulder --;
              servoShoulder.write(nShoulder);
            }
    
            if(nHead <= HEAD_ANGLE_INIT)
            {
              nHead ++;
              servoHead.write(nHead);
            }
    
            if(nJuniorHead != 70)
            {
              nJuniorHead --;
              servoJuniorHead.write(nJuniorHead);  
            }
  
            // Delay.
            
            delayMicroseconds(10000);
          }
        }
      }      
        
    // Back to home.
  
    BackToHome();
    
    // Delay.
    
    delay(2000);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Utilities.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// BackToHome()
// Move all servos to their initial positions.
// Entry   : nothing
// Returns : nothing
//

void  BackToHome()
{
  Serial.print("Back to home.\n\n");
  
  // Calculate the directions.

  nLiftDirection = (nLift >= LIFT_ANGLE_INIT) ? -1 : 1;
  nHeadDirection = (nHead >= HEAD_ANGLE_INIT) ? -1 : 1;
  nHipDirection = (nHip >= HIP_ANGLE_INIT) ? -1 : 1;
  nShoulderDirection = (nShoulder >= SHOULDER_ANGLE_INIT) ? -1 : 1;
  nJuniorArmDirection = (nJuniorArm >= JUNIOR_ARM_ANGLE_INIT) ? -1 : 1;
  nJuniorHeadDirection = (nJuniorHead >= JUNIOR_HEAD_ANGLE_END) ? -1 : 1;
  
  // Then move to home.
  
  while((nHead != HEAD_ANGLE_INIT) ||
        (nHip != HIP_ANGLE_INIT) || 
        (nLift != LIFT_ANGLE_INIT) || 
        (nShoulder != SHOULDER_ANGLE_INIT) || 
        (nJuniorArm != JUNIOR_ARM_ANGLE_INIT) ||
        (nJuniorHead != JUNIOR_HEAD_ANGLE_END))
  {
    // Head to home.
    
    if(nHead != HEAD_ANGLE_INIT)
    {
      nHead += nHeadDirection;
      servoHead.write(nHead);
    }

    // Hip to home.
    
    if(nHip != HIP_ANGLE_INIT)
    {
      nHip += nHipDirection;
      servoHip.write(nHip);
    }

    // Lift to home.
    
    if(nLift != LIFT_ANGLE_INIT)
    {
      nLift += nLiftDirection;
      servoLift.write(nLift); 
    }

    // Shoulder to home.
    
    if(nShoulder != SHOULDER_ANGLE_INIT)
    {
      nShoulder += nShoulderDirection;
      servoShoulder.write(nShoulder); 
    }

    // Junior arm to home.
    
    if(nJuniorArm != JUNIOR_ARM_ANGLE_INIT)
    {
      nJuniorArm += nJuniorArmDirection;
      servoJuniorArm.write(nJuniorArm);
    }

    // Junior head to home.
    
    if(nJuniorHead != JUNIOR_HEAD_ANGLE_END)
    {
      nJuniorHead += nJuniorHeadDirection;
      servoJuniorHead.write(nJuniorHead);
    }

    // Delay.
    
    delayMicroseconds(20000);
  }
}

//
// ForwardSwing()
// Forward swing at the ball.
// Entry   : nothing
// Returns : nothing
//

void  ForwardSwing()
{
  // Forward swing to putt the ball.
  
  Serial.print("  Swing.\n");
  
  while(nShoulder <= SHOULDER_ANGLE_INIT + SHOULDER_ANGLE_MAX - 40) // "-40" shortens the stroke
  {
    nShoulder ++;
    servoShoulder.write(nShoulder);
    delayMicroseconds(1750);
  }
}

//
// HeadsFollowBall()
// Heads follow the ball to the hole.
// Entry   : nothing
// Returns : nothing
//

void  HeadsFollowBall()
{
  // Rotate the heads to follow the ball during a putt.
  
  Serial.print("  Heads follows ball.\n");
  while((nHead >= HEAD_ANGLE_INIT - HEAD_ANGLE_MAX + 10) || (nJuniorHead < JUNIOR_HEAD_ANGLE_INIT + 50 + 20))
  {
    // Senior head.
    
    if(nHead >= HEAD_ANGLE_INIT - HEAD_ANGLE_MAX + 10)
    {
      nHead --;
      servoHead.write(nHead);
    }

    // Junior head.
    
    if(nJuniorHead < JUNIOR_HEAD_ANGLE_INIT + 50 + 20)
    {
      nJuniorHead ++;
      servoJuniorHead.write(nJuniorHead);
    }

    // Delay.
    
    delayMicroseconds(4000);
  }      
  
}

//
// PrepareToPutt()
// Prepare to putt.
// Entry   : nothing
// Returns : nothing
//

void  PrepareToPutt()
{ 
  // Junior looks to senior.

  while(nJuniorHead > 40)
  {
    nJuniorHead --;
    servoJuniorHead.write(nJuniorHead);
    delayMicroseconds(6000);
  }

  // Delay.
  
  delay(1000);
  
  // Senior bends down while junior head follows at senior.

  Serial.print("Assume the putting stance.\n");
  while((nHip <= HIP_ANGLE_INIT + HIP_ANGLE_MAX) || (nJuniorHead <= 70))
  {
    // Senior hip motion.
    
    if(nHip <= HIP_ANGLE_INIT + HIP_ANGLE_MAX)
    {
      nHip ++;
      servoHip.write(nHip);
    }

    // Junior head motion.
    
    if(nJuniorHead <= 70)
    {
      nJuniorHead ++;
      servoJuniorHead.write(nJuniorHead);
    }

    // Delay.
    
    delayMicroseconds(10000);
  }

  // Take a few practice swings.

  Serial.print("Take a few practice swings.\n");
  for(int nTemp = 0; nTemp < 4; nTemp ++)
  {
    // Forward swing.
    
    Serial.print("  Forward swing.\n");
    while(nShoulder <= SHOULDER_ANGLE_INIT + SHOULDER_ANGLE_MAX - 20)
    {
      nShoulder ++;
      servoShoulder.write(nShoulder);
      delayMicroseconds(4000);
    }

    Serial.print("  Backward swing.\n");
    while(nShoulder >= SHOULDER_ANGLE_INIT - SHOULDER_ANGLE_MAX + 20)
    {
      nShoulder --;
      servoShoulder.write(nShoulder);
      delayMicroseconds(4000);
    }
  }

  // Prepare for the ball.

  Serial.print("Prepare for the ball.\n");
  while(nShoulder <= SHOULDER_ANGLE_INIT - 10)
  {
    nShoulder ++;
    servoShoulder.write(nShoulder);
    delayMicroseconds(20000);
  }

  // Raise the ball.
  
  Serial.print("Raise the ball.\n");
  while(nLift <= LIFT_ANGLE_INIT + LIFT_ANGLE_MAX - 50)
  {
    nLift ++;
    servoLift.write(nLift);
    delayMicroseconds(5000);
  }
    
  // Eyeball the putt.
  
    Serial.print("Eyeball the putt...\n");
    for(int nTemp = 0; nTemp < 3; nTemp ++)
    {
      // Look towards hole.
      
     Serial.print("  Look towards the hole.\n");
     while(nHead >= HEAD_ANGLE_INIT - HEAD_ANGLE_MAX + 10)
      {
        nHead --;
        servoHead.write(nHead);
        delayMicroseconds(15000);
      }
  
      // Look towards ball.
      
     Serial.print("  Look towards the ball.\n");
     while(nHead <= HEAD_ANGLE_INIT)
      {
        nHead ++;
        servoHead.write(nHead);
        delayMicroseconds(15000);
      }
    }
}

//
// Putt()
// Perform the putt.
// Entry   : nothing
// Returns : nothing
//

void  Putt()
{
  // As senior backswings, junior raises flag.

  Serial.print("  Back swing.\n");
  while((nShoulder != SHOULDER_ANGLE_INIT - SHOULDER_ANGLE_MAX) || (nJuniorArm != JUNIOR_ARM_ANGLE_MAX))
  {
   if(nShoulder != SHOULDER_ANGLE_INIT - SHOULDER_ANGLE_MAX)
   {
    nShoulder --;
    servoShoulder.write(nShoulder);
   }

   if(nJuniorArm != JUNIOR_ARM_ANGLE_MAX)
   {
      nJuniorArm ++;
      servoJuniorArm.write(nJuniorArm);
   }

    delayMicroseconds(15000);
  }

  // Forward swing.

  ForwardSwing();

  // Move the heads to follow the putt.

  HeadsFollowBall();
}
